/*
 * Copyright (C) 2013 by the University of Southern California
 * $Id: utility.cc 23642 2015-04-17 23:42:34Z liangzhu $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "utility.h"
#include "types.h"

//split a string to an vector;
void Splitstring(char* cStr, char* cDelim, vector<string> &sItemVec)
{
    char* p;
    p=strtok(cStr, cDelim);
    while (p!=NULL)
    {
        sItemVec.push_back(p);
        p=strtok(NULL, cDelim);
    }
}


//compute median of an array
double compute_median(vector<unsigned long>& arr)
{
    double median;
    size_t size = arr.size();
    sort(arr.begin(), arr.end());
    if (size  % 2 == 0)
    {
        median = (arr[size / 2 - 1] + arr[size / 2]) / 2;
    }
    else
    {
        median = arr[size / 2];
    }
    return median;
}


//compute the mean of an array
double compute_mean(vector<unsigned long>& arr)
{
    size_t size = arr.size();
    double sum = accumulate(arr.begin(), arr.end(), 0.0);
    double mean = sum / size;
    return mean;
}

/*
 * This will convert www.google.com to 3www6google3com 
 * */
//function borrowed from "http://www.binarytides.com/dns-query-code-in-c-with-linux-sockets/", Author : Silver Moon
void ChangetoDnsNameFormat(unsigned char* dns,unsigned char* host)
{
    int lock = 0 , i;
    unsigned char tmp[MAX_BUF_SIZE];
    memset(tmp, 0, MAX_BUF_SIZE);
    strcpy((char*)tmp,(char*)host);
    strcat((char*)tmp,".");
    for(i = 0 ; i < strlen((char*)tmp) ; i++)
    {
        if(tmp[i]=='.')
        {
            *dns++ = i-lock;
            for(;lock<i;lock++)
            {
                *dns++=tmp[lock];
            }
            lock++; //or lock=i+1;
        }
    }
    *dns++='\0';
}

unsigned short format_edns_query(unsigned char* buf, unsigned char* host, int qtype, int qclass, int type, unsigned short msgid)
{
    unsigned short qlen=0;
    unsigned short len= format_dns_query(buf, qlen, host, qtype, qclass, type, msgid);
    unsigned char* buf_init_pos = (unsigned char*)buf;
    if(type==QT_TCP)
    {
        buf_init_pos += sizeof(unsigned short);
    }

    int ends_len = sizeof(struct EDNS_HEADER);
    //we will set the second bit of Z field to indicate "enabling TLS"
    //the most significant bit of the Z field has been used for DNSSEC ["DNSSEC OK" (DO) bit]
    unsigned char * zp =(unsigned char*)(buf + len - ends_len + ZBIT_OFFSET - 1);
    u_int16_t z_old = ns_get16(zp);
    u_int16_t z_new = z_old;
    z_new |= DNS_MESSAGEEXTFLAG_TLS;
    ns_put16(z_new, zp);

    unsigned short query_len = len + ends_len -2;
    if (type==QT_TCP)
    {
        u_int16_t len_to_write = htons(query_len);//this is very important
        memcpy(buf, &len_to_write, sizeof(len_to_write));
        query_len += sizeof(len_to_write);
    }


    return query_len;

}

//format a DNS query
//function borrowed from "http://www.binarytides.com/dns-query-code-in-c-with-linux-sockets/", Author : Silver Moon
unsigned short format_dns_query(unsigned char* buf, unsigned short &qlen, unsigned char* host, int query_type, int query_class, int type, unsigned short msgid)
{
    unsigned char *qname,*reader;
    struct DNS_HEADER *dns = NULL;
    struct QUESTION *qinfo = NULL;    
    struct EDNS_HEADER *edns_header = NULL;

    //Set the DNS structure to standard queries
    unsigned short dh_len=0;
    memset(buf, 0, MAX_BUF_SIZE);
    unsigned char* buf_init_pos = (unsigned char*)buf;

    if(type==QT_TCP)
    {
        buf_init_pos += sizeof(dh_len);
    }

    dns = (struct DNS_HEADER *)buf_init_pos;
    if(msgid==0)
        dns->id = (unsigned short) htons(getpid());
    else
        dns->id = htons(msgid);
    dns->qr = 0; //This is a query
    dns->opcode = 0; //This is a standard query
    dns->aa = 0; //Not Authoritative
    dns->tc = 0; //This message is not truncated
    dns->rd = 1; //Recursion Desired
    dns->ra = 0; //Recursion not available
    dns->z = 0;
    dns->ad = 1;
    dns->cd = 0;
    dns->rcode = 0;
    dns->q_count = htons(1); //we have only 1 question
    dns->ans_count = 0;
    dns->auth_count = 0;
    dns->add_count = htons(1); // and we now also have an EDNS0 option.
    //point to the query portion
    qname =(unsigned char*)(buf_init_pos+sizeof(struct DNS_HEADER));
    ChangetoDnsNameFormat(qname , host);
    qinfo =(struct QUESTION*)(buf_init_pos+sizeof(struct DNS_HEADER) + (strlen((const char*)qname) + 1)); //fill it

    qinfo->qtype = htons( query_type ); //type of the query , A , MX , CNAME , NS etc
    qinfo->qclass = htons(query_class); //its internet

    qlen=strlen((char*)qname)+1;
   
    //get the length of the packet;
    unsigned short len=sizeof(struct DNS_HEADER) + (strlen((const char*)qname)+1) + sizeof(struct QUESTION);

    // And stuff in an EDNS0 header after the end of the packet.
    unsigned char* buf_edns_pos = (unsigned char*)buf + len;

    if(type==QT_TCP)
      {
        buf_edns_pos += sizeof(dh_len);
      }
    edns_header = (struct EDNS_HEADER *) (unsigned char*)buf_edns_pos;
    edns_header->name = 0x0;       // EDNS0 has no name. Poor EDNS0.
    edns_header->type = htons(41); // OPT
    edns_header->udp_payload = htons(UDP_ANS_SIZE);
    edns_header->rcode = 0x0;      // Extended RCODE
    edns_header->version = 0x0;    // EDNS*0*.
    edns_header->Z = htons(DNS_MESSAGEEXTFLAG_DO);// Flags.
    edns_header->len = 0x0;        // EDNS0 still has no name...

    // Hey, the length got bigger!
    len += sizeof(struct EDNS_HEADER);
     
    if(type==QT_TCP)
    {
        dh_len = htons(len);
        memcpy(buf, &dh_len, sizeof(dh_len));
        return len+sizeof(dh_len);
    }
    else
    {
        return len;
    }
}

void print_section(ns_msg handle, int count, ns_sect sec_t, const char* sec_name)
{
    ns_rr rr;
    char buf[MAX_BUF_SIZE];
    cout << sec_name << ":" << endl;
    for (int rrnum=0; rrnum<count; rrnum++) 
    {
        memset(&rr, 0, sizeof(rr));
        if (ns_parserr(&handle, sec_t, rrnum, &rr)) 
        {
            cerr<<"[ERROR]ns_parserr"<<endl;
            exit(1);
        }
        memset(buf, '\0', sizeof(buf));
        ns_sprintrr (&handle, &rr, NULL, NULL, buf, sizeof (buf));
        cout << buf << endl;
    }
}

unsigned short get_dns_msg_id(unsigned char* msg, int msg_len)
{
    ns_msg  handle;
    if (ns_initparse(msg, msg_len, &handle) < 0) {
	cerr << "failed to parse msg with length " << msg_len << endl;
	perror("ns_initparse");
        exit(1);
    }

    u_int16_t ID      = ns_msg_id(handle);
    u_int16_t QR      = ns_msg_getflag(handle, ns_f_qr);
    u_int16_t AA      = ns_msg_getflag(handle, ns_f_aa);
    u_int16_t TC      = ns_msg_getflag(handle, ns_f_tc);
    u_int16_t RD      = ns_msg_getflag(handle, ns_f_rd);
    u_int16_t RA      = ns_msg_getflag(handle, ns_f_ra);
    u_int16_t Z       = ns_msg_getflag(handle, ns_f_z);
    u_int16_t Opcode  = ns_msg_getflag(handle, ns_f_opcode);
    u_int16_t Rcode   = ns_msg_getflag(handle, ns_f_rcode);
    u_int16_t QdCount = ns_msg_count(handle, ns_s_qd);
    u_int16_t AnCount = ns_msg_count(handle, ns_s_an);
    u_int16_t NsCount = ns_msg_count(handle, ns_s_ns);
    u_int16_t ArCount = ns_msg_count(handle, ns_s_ar);
    return ID;
}


void print_dns_msg(unsigned char* msg, int msg_len)
{
    ns_msg  handle;
    if (ns_initparse(msg, msg_len, &handle) < 0) 
    {
        perror("ns_initparse");
        exit(1);
    }

    u_int16_t ID      = ns_msg_id(handle);
    u_int16_t QR      = ns_msg_getflag(handle, ns_f_qr);
    u_int16_t AA      = ns_msg_getflag(handle, ns_f_aa);
    u_int16_t TC      = ns_msg_getflag(handle, ns_f_tc);
    u_int16_t RD      = ns_msg_getflag(handle, ns_f_rd);
    u_int16_t RA      = ns_msg_getflag(handle, ns_f_ra);
    u_int16_t Z       = ns_msg_getflag(handle, ns_f_z);
    u_int16_t Opcode  = ns_msg_getflag(handle, ns_f_opcode);
    u_int16_t Rcode   = ns_msg_getflag(handle, ns_f_rcode);
    u_int16_t QdCount = ns_msg_count(handle, ns_s_qd);
    u_int16_t AnCount = ns_msg_count(handle, ns_s_an);
    u_int16_t NsCount = ns_msg_count(handle, ns_s_ns);
    u_int16_t ArCount = ns_msg_count(handle, ns_s_ar);
    cout << "-----DNS Msg: len=" << msg_len << "; "<< (QR==0 ? "query" : "response") << "; id="<< ID <<"-----" << endl;
    print_section(handle, QdCount, ns_s_qd, "Que");
    print_section(handle, AnCount, ns_s_an, "Ans");
    print_section(handle, NsCount, ns_s_ns, "Aus");
    print_section(handle, ArCount, ns_s_ar, "Add");
}

bool read_dnslist(char* fn, vector<string>& dns_list)
{
    char buf[MAX_BUF_SIZE];
    FILE*fp = fopen(fn, "r");
    if(!fp)
        return false;
    while (fgets(buf, MAX_BUF_SIZE, fp))
    {
        buf[strlen(buf)-1]='\0';
        if(buf[0] == '#')
            continue;
        dns_list.push_back(buf);
        memset(buf, 0, MAX_BUF_SIZE);
    }
    fclose(fp);
    return true;
}

void generate_uniq_random(vector<string> &dns_list, int num)
{
    srand (time(NULL));
    set<string> uniq_dns;
    if(num>dns_list.size())
    {
        printf("exceed max: %d\t%lu\n", num, dns_list.size());
    }
    int i=0;
    for(i=0; i<num; )
    {
        int rd_index = rand()%(dns_list.size());
        if(uniq_dns.find(dns_list[rd_index])!= uniq_dns.end())
        {
            continue;
        }
        else
        {
            printf("%s\n", (dns_list[rd_index]).c_str());
            uniq_dns.insert(dns_list[rd_index]);
            i++;
        }
    }
}

bool is_tls_set (unsigned char* msg, int msg_len)
{
    ns_msg  handle;
    if (ns_initparse(msg, msg_len, &handle) < 0) 
    {
        perror("ns_initparse");
        exit(1);
    }

    u_int16_t ArCount = ns_msg_count(handle, ns_s_ar);
    if (ArCount == 0)//no OPT
        return false;
    ns_rr rr;
    //find the first query name: this should be 'STARTTLS' 
    memset(&rr, 0, sizeof(rr));
    if (ns_parserr(&handle, ns_s_qd, 0, &rr))
    {
        perror("ns_parserr");
    }

    string rr_name = ns_rr_name(rr);
    int q_type = ntohs(rr.type);
    int q_class = ntohs(rr.rr_class);
    //printf("qname: %s, TYPE :%d, CLASS:%d\n", rr_name.c_str(),  q_type, q_class);

    unsigned char* p = (unsigned char*)(msg + sizeof(struct DNS_HEADER));
    ns_name_skip((const unsigned char**)&p, p+(msg_len - sizeof(struct DNS_HEADER)));
    p=p+sizeof(struct QUESTION);

    u_int16_t *t = (u_int16_t *)(p+1);
    u_int16_t rr_type = ntohs(*t);
    u_int16_t rr_class = ntohs(*(t+1));
    u_int16_t rr_ttl= ntohs(*(t+3));
    //printf("TYPE: %04x, CLASS:%04x,  TTL bits:%04x\n", rr_type, rr_class, rr_ttl);
    if((rr_ttl & DNS_MESSAGEEXTFLAG_TLS) || (strcmp (rr_name.c_str(), "STARTTLS" )==0))
    {
        //tls bit is set
        return true;
    }
    return false;
}
